package coreElements;

import gameElements.Blocker;
import gameElements.Character;
import gameElements.GridCell;
import gameElements.Racer;

import java.awt.*;
import java.awt.image.BufferStrategy;
import java.util.Iterator;

import javax.swing.JApplet;

/**
 * @author      Christopher Dragert
 * 
 * The applet container.  Manages the GUI, launches the game.  Uses double-buffered graphics for smooth drawing without
 * flicker.
 * 
 */
public class FBMain extends JApplet {
	
	/**
	 * 
	 */
	private static final long serialVersionUID = -4141188207429021055L;
	
	private Game game;

	private boolean showBraids;
	private boolean showVelocity;
	private PropertiesManager properties;
	
	//gui components
	private Canvas drawArea;
	private MenuPanel menuPanel;
	
	//listeners
	private ExitListener mouseListener;
	private MenuListener keyListener;
	
	private boolean menuActive;
	
	public void setShowBraids(boolean showBraids) {
		this.showBraids = showBraids;
	}

	public void setShowVelocity(boolean showVelocity) {
		this.showVelocity = showVelocity;
	}
	
	public void init() {
		this.start();
		this.setBackground(Color.gray);
		this.setPreferredSize(new Dimension(Game.WIDTH+70, Game.HEIGHT+90));
				
		properties = new PropertiesManager(this);
		properties.applyProperties();
		
		drawArea=new Canvas();
		drawArea.setSize(Game.WIDTH+70, Game.HEIGHT+90);
		
		this.getContentPane().add("Center", drawArea);
		drawArea.setVisible(true);
		
		//use frame buffering for smooth drawing
		drawArea.createBufferStrategy(2);
		
		//add listeners
		mouseListener = new ExitListener();
		keyListener = new MenuListener(this);
		drawArea.addMouseListener(mouseListener);
		drawArea.addKeyListener(keyListener);
		
		//set it to take focus
		drawArea.setFocusable(true);
		drawArea.requestFocus();
		
		//now validate and resize the applet
		this.revalidate();
		this.resize(this.getPreferredSize());
		
		//prepare the menu panel
		menuPanel = new MenuPanel(properties, this);
		menuActive = false;
		
		//start game thread
		game = Game.getInstance();
		game.init(this);
		Thread gameThread = new Thread(game);
		gameThread.setPriority(Thread.MIN_PRIORITY);
		gameThread.start(); // start Game Loop.
		
	}
	
	public void update(Graphics g) {
		if (menuActive == false) {
			paint(g);
		}
	}
	
	public void paint(Graphics g){
		if (menuActive == false) {
			BufferStrategy bf = drawArea.getBufferStrategy();
	
			Graphics bufferGraphics = bf.getDrawGraphics();
			paintFrame(bufferGraphics);
			bufferGraphics.dispose();
			bf.show();
			
			paintFrame(g);
			Toolkit.getDefaultToolkit().sync();
		}
		else {
			this.paintComponents(g);
		}
	}
	
	public void paintFrame(Graphics g) {
		Graphics2D g2 = (Graphics2D)g;
		
		//redraw gray border:
		g2.setColor(Color.gray);
		g2.fillRect(0,0,Game.WIDTH+Game.BUFFER*2, Game.BUFFER);
		g2.fillRect(0,Game.BUFFER,Game.BUFFER, Game.HEIGHT+Game.BUFFER);
		g2.fillRect(0,Game.HEIGHT+Game.BUFFER,Game.WIDTH+Game.BUFFER*2, Game.HEIGHT+Game.BUFFER*2);
		g2.fillRect(Game.WIDTH+Game.BUFFER,Game.BUFFER,Game.WIDTH+Game.BUFFER*2, Game.HEIGHT+Game.BUFFER);
		
		//draw water
		g2.setColor(Color.blue);
		g2.fillRect(Game.BUFFER, Game.BUFFER, Game.WIDTH, Game.HEIGHT);
		
		//draw game elements
		paintMaze(g2);
		paintCharacters(g2);
		paintFPS(g2);
		paintString(g2);
	}
	
	public void paintFPS(Graphics2D g2) {
		g2.setColor(Color.white);
		int actualFPS = game.getActualFPS();
		int desiredFPS = game.getDesiredFPS();
		if (actualFPS < desiredFPS)
			g2.setColor(Color.red);	
		g2.drawString("FPS:" + actualFPS , Game.BUFFER, Game.HEIGHT + Game.BUFFER*2 -5 );
	}
	
	public void paintString(Graphics2D g2) {
		g2.setColor(Color.white);
		g2.drawString("Press (m) to change settings, click mouse on grid to move exit", 600, Game.HEIGHT + Game.BUFFER*2 -5 );
	}
	
	public void paintMaze(Graphics2D g2) {
		g2.setColor(Color.green);
		for (int i = 0;i<game.getMaze().size+1;i++){
			for (int j = 0;j<game.getMaze().size;j++){
				GridCell gc = game.getMaze().getCells()[i][j];
				if (gc.braided && showBraids) {
					g2.setColor(Color.orange);
				}
				if (gc.inMaze())
					g2.fillRect(Game.BUFFER + gc.guiX, Game.BUFFER + gc.guiY,
							gc.width, gc.height);
				if (gc.braided && showBraids) {
					g2.setColor(Color.green);
				}
				if (game.getMaze().getCells()[i][j].isExit()) {
					game.getMaze().exit = game.getMaze().getCells()[i][j];
				}
			}
		}
		g2.setColor(Color.black);
		g2.drawString("EXIT", Game.BUFFER + game.getMaze().exit.guiX + 5, Game.BUFFER+game.getMaze().exit.guiY + 13);
		g2.drawRect(Game.BUFFER + game.getMaze().exit.guiX, Game.BUFFER + game.getMaze().exit.guiY,
				game.getMaze().exit.width, game.getMaze().exit.height);
	}
	
	public void paintCharacters(Graphics2D g2) {
		
		if (!game.getCharacters().isEmpty()) {
			for (Iterator<Character> i = game.getCharacters().iterator(); i.hasNext();){
				Character c = i.next();
				if (c instanceof Racer) {
					g2.setColor(Color.red);
					g2.fillOval((int)c.p.x + Game.BUFFER - game.RADIUS, (int)c.p.y + Game.BUFFER - game.RADIUS, 10, 10);
				}
				if (c instanceof Blocker) {
					g2.setColor(Color.black);
					g2.fillOval((int)((Blocker)c).p.x + Game.BUFFER - game.RADIUS, (int)((Blocker)c).p.y + Game.BUFFER - game.RADIUS, 10, 10);
				}
				if (showVelocity) {
					drawDirections(g2,c);
				}
			}
		}
	}
	
	public void drawDirections(Graphics2D g2, Character c) {
		g2.setColor(Color.black);
		int a = Game.BUFFER;
		g2.drawLine((int)(c.p.x+a) , (int)(c.p.y+a), (int)(c.p.x+c.v.x*10+a), (int)(c.p.y+c.v.y*10+a));
		//g2.drawLine((int)(c.p.x+a) , (int)(c.p.y+a), (int)(c.p.x+c.F.x*10+a), (int)(c.p.y+c.F.y*10+a));
	}
	
	public void openMenu() {
		
		menuActive = true;
		
		//drawArea.removeKeyListener(keyListener);
		//drawArea.removeMouseListener(mouseListener);
		//drawArea.setVisible(false);
		game.stopGame();
		
		this.getContentPane().remove(drawArea);
		
		//createMenu
		menuPanel.readValues();
		this.getContentPane().add("Center", menuPanel);
		menuPanel.setVisible(true);	
		
		this.revalidate();
		this.repaint();
		
		
		this.resize(this.getPreferredSize());
		
	}
	
	public void startNewGame() {
		//start game with new settings
		properties.applyProperties();
		game = Game.getInstance();
		game.init(this);

		menuPanel.setVisible(false);
		menuActive = false;
		
		this.getContentPane().add("Center", drawArea);
		drawArea.setVisible(true);
		drawArea.createBufferStrategy(2);
		drawArea.requestFocus();
		
		this.revalidate();
		this.repaint();
		
		//launch game thread
		Thread gameThread = new Thread(game);
		gameThread.setPriority(Thread.MIN_PRIORITY);
		gameThread.start(); // start Game Loop.
		
	}
	
}



