package coreElements;

import gameElements.Blocker;
import gameElements.Character;
import gameElements.GridCell;
import gameElements.Maze;
import gameElements.Racer;

import java.util.Iterator;
import java.util.Stack;
import java.util.Vector;
import java.util.concurrent.Semaphore;


/**
 * @author      Christopher Dragert
 * 
 * The core game thread.  Contains the game loop, applies game settings, creates the game, applies game logic, and so on. 
 * 
 */
public final class Game implements Runnable {
	
	FBMain app;
	
	private static Game game;
	
	private Semaphore mutex;
	
	//stuff for tracking FPS
	static int desiredFPS;
	private double milliFPS;
	private int actualFPS;
	
	public static final int WIDTH = 1024;
	public static final int HEIGHT = 768;
	public static final int BUFFER = 25;
	
	//the radius of the characters
	public final int RADIUS = 5;
	

	//the relative scale from n to screen size.
	public static double x_scale;
	public static double y_scale;
	public double diag;

	static int part;
	static int numRacers;
	static boolean racerRespawn;
	static int numBlockers;
	static boolean blockersMove;
	
	static int size;
	static int density;
	static int braids;
	
	public static void setProperties(int part, int desiredFPS, int numRacers, boolean racerRespawn, int numBlockers, boolean blockersMove,
			int size, int density, int braids) {
		Game.part = part;
		Game.desiredFPS = desiredFPS;
		Game.numRacers = numRacers;
		Game.racerRespawn = racerRespawn;
		Game.numBlockers = numBlockers;
		Game.blockersMove = blockersMove;
		Game.size = size;
		Game.density = density;
		Game.braids = braids;
	}
	
	//game element holders
	public Vector<Character> characters;
	//the maze.  each cell is either in the maze or out of it.
	private Maze maze;

	public final static Game getInstance() {
		if (game==null) {
			game = new Game();
		}
		return game;
	}
	
	private Game() {
		mutex = new Semaphore(0);
	}
	
	
	boolean running;
	
	public void init(FBMain app){

		this.app = app;
		milliFPS = 1000/desiredFPS;
		maze = new Maze(Game.size, Game.density, Game.braids);
		
		characters = new Vector<Character>();
		
		Blocker.move = blockersMove;
		
		if (part == 1){
			//nothing else to do.
		}
		else if (part==2){
			//create racers
			createRacers(numRacers);
		}
		else if (part==3){ 
			createRacers(numRacers);
			createBlockers(numBlockers);
		}
		else {
			System.out.println("Invalid property.  Set 'part' to one of [1,2,3].");
			System.exit(0);
		}
		
		running = true;
	
	}

	@Override
	public void run() {
		long frameStartTime;
		long frameTime = 0;
		
		//the game loop
		while (running) {

			frameStartTime = System.currentTimeMillis();
			if (part > 1) {
				for (Iterator<Character> i = characters.iterator();i.hasNext();) {
					Character c = i.next();
					if (c instanceof Blocker && Blocker.move || !(c instanceof Blocker)) {
						c.clearForces();
						applyGridCollision(c);			
						c.applyGoals();
						checkCollisions(c);
						c.update(1);
					}
				}
				if (pendingRemoval != null) {
					characters.remove(pendingRemoval);
					if (racerRespawn)
						createRacers(1);
					pendingRemoval = null;
				}

			}
			app.repaint();
			//app.paint(app.getGraphics());
			//app.update(app.getGraphics());
			
			frameTime = System.currentTimeMillis() - frameStartTime;
			synchronized(this) {
				if (frameTime < milliFPS) {
					try {
						Thread.sleep((long)milliFPS-frameTime );
					}
					catch (Exception e) {
						//woken up - not important
					}
					actualFPS = desiredFPS;
				}
				else {
					actualFPS = (int)(1000/frameTime);
				}
			}
		}	
		mutex.release();
	}
	
	private Racer pendingRemoval = null;
	/*
	 * remove this character from the game.
	 * Typically called when a racer reaches the exit!
	 */
	public void remove(Racer r) {
		pendingRemoval = r;
	}
	
	private GridCell g;
	private GridCell g2;
	private double nearness;
	
	public void applyGridCollision(Character c) {
		//perform collision detection with maze boundaries
		//find neighbours, repulse character if neighbour is not in maze.
		g = maze.getGridLocation(c.p);
		
		//left column
		if (g.x>0) {
			//top left
			if (g.y > 0) {
				g2 = maze.getCells()[g.x-1][g.y-1];
				if (!g2.inMaze()) {
					nearness = Math.sqrt((c.p.x - g.guiX)*(c.p.x - g.guiX) + (c.p.y - g.guiY)*(c.p.y - g.guiY));
					if (nearness < y_scale*.1) {
						c.addForce(400/nearness,400/nearness);
					}
				}
			}
			
			// middle left
			g2 = maze.getCells()[g.x-1][g.y];
			if (!g2.inMaze()) {
				nearness = c.p.x-g2.guiX-x_scale;
				if (nearness < x_scale*.15) {
					c.addForce(400/nearness, 0);
				}
			}
			
			// bottom left
			if (g.y < maze.size - 1) {
				g2 = maze.getCells()[g.x-1][g.y+1];
				if (!g2.inMaze()) {
					nearness = Math.sqrt((c.p.x - g.guiX)*(c.p.x - g.guiX) + (c.p.y - (g.guiY + y_scale))*(c.p.y - (g.guiY + y_scale)));
					if (nearness < y_scale*.1) {
						c.addForce(400/nearness,-400/nearness);
					}
				}
			}
		}
		
		//top cell
		if (g.y>0) {
			g2 = maze.getCells()[g.x][g.y-1];
			if (!g2.inMaze()) {
				nearness = c.p.y-g2.guiY-y_scale;
				if (nearness < y_scale*.15) {
					c.addForce(0, 400/nearness);
				}
			}						
		}
		//bottom cell
		if (g.y< maze.size -1) {
			g2 = maze.getCells()[g.x][g.y+1];
			if (!g2.inMaze()) {
				nearness = g2.guiY - c.p.y;
				if (nearness < y_scale *.15) {
					c.addForce(0, -400/nearness);
				}
			}			
		}
		
		if (g.x< maze.size ) {
			//top right
			if (g.y > 0) {
				g2 = maze.getCells()[g.x+1][g.y-1];
				if (!g2.inMaze()) {
					nearness = Math.sqrt((c.p.x - (g.guiX + x_scale))*(c.p.x - (g.guiX + x_scale)) + (c.p.y - g.guiY)*(c.p.y - g.guiY));
					if (nearness < y_scale*.1) {
						c.addForce(-400/nearness,400/nearness);
					}
				}
			}
			
			//middle right
			if (g.x< maze.size ) {
				g2 = maze.getCells()[g.x+1][g.y];
				if (!g2.inMaze()) {
					nearness = g2.guiX - c.p.x;
					if (nearness < x_scale *0.15) {
						c.addForce(-400/nearness, 0);
						//System.out.print(-0.2*x_scale-nearness/(0.2*x_scale));
					}
				}			
			}
			
			// bottom left
			if (g.y <  maze.size  - 1) {
				g2 = maze.getCells()[g.x+1][g.y+1];
				if (!g2.inMaze()) {
					nearness = Math.sqrt((c.p.x - (g.guiX + x_scale))*(c.p.x - (g.guiX + x_scale)) + (c.p.y - (g.guiY + y_scale))* (c.p.y - (g.guiY + y_scale)));
					if (nearness < y_scale*.1) {
						c.addForce(-400/nearness,-400/nearness);
					}
				}
			}
		}
		
		if (c.p.x < x_scale/10) {
			//repel by the wall
			c.addForce(400 * (x_scale/4 - c.p.x), 0.0);
		}
		if (c.p.x > WIDTH - x_scale/10) {
			c.addForce(400 * (c.p.x - WIDTH), 0.0);
		}
		
		if (c.p.y < y_scale/10) {
			//repel by the wall
			c.addForce(0.0, 400 * (y_scale/4 - c.p.y));
		}
		if (c.p.y > HEIGHT - y_scale/10) {
			c.addForce(0.0, 400 * (c.p.y - HEIGHT));
		}
	}
	
	private Character c2;
	private Vector2d v = new Vector2d(0,0);
	private double scale;
	
	/*
	 * Check a character for collisions with all other characters
	 *  (No smart strategy here)
	 *  
	 *  Collision rules:
	 *  	- Racers push Racers
	 *  	- Blockers push Racers
	 *  	- Racers push blockers, but only 10% as much as the inverse
	 *  	- Blockers push Blockers
	 * 
	 */
	private void checkCollisions(Character c) {
		
		if (c instanceof Racer) {
		
			for (Iterator<Character> i2 = characters.iterator();i2.hasNext();){
				c2 = i2.next();
				
				//make sure they're not the same object
				if (c!=c2) {
					//if they within 2*RADIUS of each other, they are colliding
					nearness = c.p.distance(c2.p);
					if (nearness < RADIUS * 3) {
						v.set(c.p.x-c2.p.x+0.01, c.p.y - c2.p.y);
						scale = 1/(nearness-RADIUS*2);
						if (scale < 0) scale = 10;
						v.scale(scale);
						c.addForce(v);
					}
				}
			}
		}
		else if (c instanceof Blocker) {
			for (Iterator<Character> i2 = characters.iterator();i2.hasNext();){
				c2 = i2.next();
				
				//make sure they're not the same object
				if (c!=c2) {
					if (Blocker.move || c2 instanceof Blocker  ) {
						//if they within 2*RADIUS of each other, they are colliding
						nearness = c.p.distance(c2.p);
						if (nearness < RADIUS * 3) {
							v.set(c.p.x-c2.p.x+0.01, c.p.y - c2.p.y);
							scale = .1/(nearness-RADIUS*2);
							if (scale < 0) scale = 1000;
							v.scale(scale);
							c.addForce(v);
						}
					
						if (c2 instanceof Racer) {
							// we have racer - need this for blocking behaviour.
							nearness = c.p.distance(c2.p);
							if (nearness < RADIUS * 50) {
								((Blocker)c).addBlockee(c2, nearness);
							}
						}
					}
				}
			}
		}
	}
	
	
	public void createRacers(int num) {		
		
		//set their max speed:
		Racer.max_v = 4 - (Math.sqrt(Math.sqrt(( maze.size ))));
		
		Stack<GridCell> path = null;
		double y_start;
		
		for (int i = 0; i<num; i++) {
		
			y_start = Math.random()* HEIGHT + RADIUS;
			if (y_start > HEIGHT-10) y_start = HEIGHT-10;
			//System.out.print("Starting pathfinding for Racer " + (i+1) +"...");
			path =  AStar.getPath(maze.getGridLocation(new Vector2d(0,y_start)), maze.exit);
			//System.out.println("done!");
			characters.add(new Racer(x_scale/2,y_start, path));
		}
	}
	
	public void createBlockers(int num) {		
		
		//set their max speed:
		Blocker.max_v = 4 - (Math.sqrt(Math.sqrt(( maze.size ))));
		
		double x_start = 0;
		double y_start = 0;
		boolean inMaze = false;
		
		for (int i = 0; i<num; i++) {
			while (!inMaze) {
				x_start = Math.random()*(WIDTH - x_scale) + x_scale + 10;
				if (x_start > WIDTH-10) x_start = WIDTH-10;
				y_start = Math.random()* HEIGHT + RADIUS;
				if (y_start > HEIGHT-10) y_start = HEIGHT-10;
				
				if (maze.getGridLocation(x_start, y_start).inMaze() ) break;
			}
			//System.out.print("Starting pathfinding for Racer " + (i+1) +"...");
			//path =  AStar.getPath(getGridLocation(new Vector2d(0,y_start)), exit);
			//System.out.println("done!");
			characters.add(new Blocker(x_start,y_start, null));
		}
	}
	
	public void notifyNewExit() {
		for (Iterator<Character> i = characters.iterator();i.hasNext();) {
			Character c = i.next();
			c.notifyNewExit();
		}
	}
	
	public void stopGame() {
		running = false;
		mutex.acquireUninterruptibly();
		game = null;
	}
	
/**
 * getters and setters
 */
	
	public synchronized Maze getMaze() {
		return maze;
	}
	public synchronized Vector<Character> getCharacters() {
		return characters;
	}
	public int getDesiredFPS() {
		return desiredFPS;
	}

	public int getActualFPS() {
		return actualFPS;
	}
	
	
	
	
}
