package gameElements;

import java.util.Iterator;
import java.util.Vector;

import coreElements.Game;
import coreElements.Vector2d;

/**
 * @author      Christopher Dragert
 * 
 * The maze.  Contains grid cells that may or may not be in the maze.  Handles all
 * maze generation.
 * 
 */
public class Maze {
	
	/*
	 * Maze variables
	 */
	//the maze size.  calling it 'n' to maintain consistency with the assignment
	public int size;
	//the maze density
	private int density;
	//the number of braids in the maze
	private int braids;

	private GridCell[][] cells;
	
	public synchronized GridCell[][] getCells() {
		return cells;
	}

	public GridCell exit;

	public Maze(int size, int density, int braids ) {
		this.size = size;
		this.density = density;
		this.braids = braids;
	
		makeMaze();
	}
	
	/*
	 * Creates an (n+1) by (n) maze.
	 */
	private void makeMaze() {	
		
		Game.x_scale =((double)Game.WIDTH)/(size+1);
		Game.y_scale =((double)Game.HEIGHT)/(size);
		
		//create grid cells
		cells = new GridCell[size+1][size];
		for (int i = 0;i<size+1;i++){
			for (int j = 0;j<size;j++){
				cells[i][j] = new GridCell(i,j,(int)Math.round(i*Game.x_scale),(int)Math.round(j*Game.y_scale),
										  (int)Math.ceil(Game.x_scale),(int)Math.ceil(Game.y_scale)+1);
			}
		}
		
		//put the first column into the maze.
		for (int j =0; j<size;j++) {
			cells[0][j].putInMaze();
		}
		
		int min = (size-5)/5;
		if (min>100) min = 100;
		//now, add new random paths based upon the density
		for (int i = 1; i <= density; i++) {
			int targetx = (int)Math.floor(Math.random()*size) + 1;
			int targety = (int)Math.floor(Math.random()*size);
			
			int attempts = 50;
			while(!randomWalk(cells[targetx][targety],0,min, null)){
				if (attempts == 0) break; 
				//that destination didn't work - pick another.
				targetx = (int)Math.floor(Math.random()*size) + 1;
				targety = (int)Math.floor(Math.random()*size);
				attempts--;
			}
		}
		
		//set the exit and connect it.
		int exitRow = (int) Math.floor(Math.random()*size);
		while (!randomWalk(cells[size][exitRow],0,0, null)) {
			exitRow = (int) Math.floor(Math.random()*size);
		}
		cells[size][exitRow].setExit(true);
		exit = cells[size][exitRow];
		Racer.setExit(exit);
		
		//maze is almost done.  need to add braids
		min = (int)Math.sqrt(min);
		int attempts = 100;
		while (braids > 0 && attempts >0) { 
			int targetx = (int)Math.floor(Math.random()*size) + 1;
			int targety = (int)Math.floor(Math.random()*size);
			if(braidedWalk(cells[targetx][targety], min)) {
				braids--;
			}
			attempts--;
		}
	}
	
	private boolean braidedWalk(GridCell gc, int min) {
		
		if (!gc.inMaze()) return false;
		
		//first build the blacklist
		Vector<GridCell> blacklist = buildBlackList(gc);

		if (randomWalk(gc, 0, min, blacklist)) return true;
		return false;
	}
	
	public Vector<GridCell> buildBlackList(GridCell gc) {
		Vector<GridCell> blacklist = new Vector<GridCell>();
		blacklist.add(gc);
		Vector<GridCell> neighbours = get4Neighbours(gc, null);
		for (Iterator<GridCell> i = neighbours.iterator(); i.hasNext();) {
			GridCell current = i.next();
			if (current.inMaze()) {
				blacklist.add(current);
			}
		}
		return blacklist;
	}
	
	private boolean randomWalk(GridCell gc, int length, int min, Vector<GridCell> blacklist) {
		
		//set pending to prevent random walk from 'doubling back'
		gc.setVisited(true);
		Vector<GridCell> neighbours = get4Neighbours(gc, blacklist);
		
		//check neighbours for connection
		if (connected(neighbours)) {
			//make sure we have met minimum:
			if (length >= min) {
				//connected!  we have completed our path.
				gc.putInMaze();
				if (blacklist != null) {
					gc.setBraided();
				}
				gc.setVisited(false);
				return true;
			}
			else {
				//Discard this path.
				gc.setVisited(false);
				return false;
			}
		}

		//now see if a move is possible
		// i.e. not blocked-in by already visited nodes.
		if (blocked(neighbours) || neighbours.size()==0) {
			//blocked!  Discard this path.
			gc.setVisited(false);
			return false;
		}
		
		//not yet connected but not blocked - recursively take another step
		GridCell nextCell = getNextCell(neighbours);	
		if (randomWalk(nextCell, length + 1, min, blacklist)) {
			gc.putInMaze();
			if (blacklist != null) {
				gc.setBraided();
			}
			return true;
		}
		else {
			gc.setVisited(false);
			return false;
		}
	}
	
	public static GridCell getNextCell(Vector<GridCell> neighbours){
		int target;
		GridCell nextCell = null;
		boolean stepped = false;
		while (!stepped) {
			//take a step
			target = (int)Math.floor(Math.random()*neighbours.size());
			nextCell = neighbours.get(target);
			if (!nextCell.visited())
				stepped = true;
		}
		return nextCell;
	}
	
	public static boolean connected(Vector<GridCell> neighbours) {
		for (Iterator<GridCell> i = neighbours.iterator(); i.hasNext();) {
			GridCell current = i.next();
			if (current.inMaze()) {
				return true;
			}
		}
		return false;
	}
	
	public static boolean blocked(Vector<GridCell> neighbours) {
		
		for (Iterator<GridCell> i = neighbours.iterator(); i.hasNext();) {
			GridCell current = i.next();
			if (!current.visited()) {
				return false;
			}
		}
		return true;
	}

	public Vector<GridCell> get4Neighbours(GridCell gc, Vector<GridCell> blacklist) {
		Vector<GridCell> neighbours = new Vector<GridCell>();
		if (gc.x>0) neighbours.add(cells[gc.x-1][gc.y]);
		if (gc.x<size) neighbours.add(cells[gc.x+1][gc.y]);
		if (gc.y>0) neighbours.add(cells[gc.x][gc.y-1]);
		if (gc.y<size-1) neighbours.add(cells[gc.x][gc.y+1]);
		if (blacklist != null) {
			for (Iterator<GridCell> i = blacklist.iterator(); i.hasNext();) {
				GridCell current = i.next();
				if (neighbours.contains(current)) {
					neighbours.remove(current);
				}
			}
		}
		return neighbours;
	}
	
	/*
	 * Takes a point, and determines which grid cell it is in.
	 */
	public GridCell getGridLocation(Vector2d v) {
		return cells[(int)(Math.floor(v.x / Game.x_scale))]
		           [(int)(Math.floor(v.y / Game.y_scale))];
	}
	
	public GridCell getGridLocation(double x, double y) {
		return cells[(int)(Math.floor(x / Game.x_scale))]
		           [(int)(Math.floor(y / Game.y_scale))];
	}
	
	public void printMaze() {
		for (int i = 0; i<size; i++) {
			for (int j = 0; j< size+1;j++) {
				if (cells[j][i].inMaze()) {
					System.out.print("X");
				}
				else System.out.print("-");
			}
			System.out.println("");
			
		}
	}
	
	public void setExit(double x, double y) {
		GridCell newExit = getGridLocation(x, y);
		if (newExit.inMaze()) {
			this.exit.setExit(false);
			newExit.setExit(true);
			Racer.setExit(newExit);
			Game.getInstance().notifyNewExit();
		}
	}
}
